#!/usr/bin/env python
#-*- coding:utf-8 -*-
#===============================================================================
#  Copyright (C) 2010 Schaefer & Tobies SuC GmbH.
#  Author: Markus Grimm <mgr@schaefer-tobies.de>
#
#  This file is part of Pyontapi.
#
#  Pyontapi is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyontapi is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Pyontapi.  If not, see <http://www.gnu.org/licenses/>.
#
#===============================================================================

"""
This module is used to invoke ONTAPI commands on a NetApp Filer, NetCache, 
Agent or DFM.

After creating a :class:`schtob.pyontapi.NaFiler` object, it 
automatically invokes the API command **system-get-ontapi-version** on the 
filer to determine the API version.

Then, if there is no module 
*schtob.pyontapi.api.ontapi_<major_version>_<minor_version>*, it generates the 
module using some API commands which are were intruduced in ONTAPI 1.6.
If you want to know which API commands are available on your filer, read the 
original NetApp documentation or instantiate the filer object for the 
first time and have a look at the generated python code. Another way would be 
running the command: 

.. code-block:: bash

    pyontapi_list_commands.py

This gives you an overview of the API commands available on certain storage 
system.

The API commands are wrapped into command packages which are added to the 
:class:`schtob.pyontapi.NaFiler` instance as instance attributes.

This example shows you how to use a filer instance::

    >>> from schtob.pyontapi import NaFiler, errors
    >>> filer = NaFiler('filer-name', {'user': 'root', 'password': 'foobar'})
    >>> try:
    ...     result = filer.volume.list_info()
    ... except errors.APIFailure, exc:
    ...     print exc.get_error()
    ... 
    >>> print result['volumes']

"""

import base64
import logging
import xml.dom.minidom

from schtob.pyontapi import api, constants, errors
from schtob.pyontapi.http_responses import httplib

# Custom log level for pyontap logger. set to logging.DEBUG for debugging
# if set to logging.NOTSET, the default logger settings are used
# visit http://docs.python.org/library/logging.html for more details.
LOG_LEVEL = logging.NOTSET

class NaFiler(object):
    
    """
    Create a new connection to filer ``filer`` using ``settings`` dict.
    
    ``settings`` may consist of:

    +--------------------+-----------+-------------------------------------+
    | Key                | Default   | Possible values                     |
    +====================+===========+=====================================+
    | **user**           | "root"    | ``str``                             |
    +--------------------+-----------+-------------------------------------+
    | **password**       | ""        | ``str``                             |
    +--------------------+-----------+-------------------------------------+
    | **style**          | ``LOGIN`` | ``LOGIN``, ``HOSTS``                |
    +--------------------+-----------+-------------------------------------+
    | **vfiler**         | ""        | ``str``                             |
    +--------------------+-----------+-------------------------------------+
    | **server_type**    | "Filer"   | "Filer", "NetCache", "DFM", "Agent" |
    +--------------------+-----------+-------------------------------------+
    | **timeout**        | 0         | ``int``                             |
    +--------------------+-----------+-------------------------------------+
    | **transport_type** | ``HTTP``  | ``HTTP``, ``HTTPS``                 |
    +--------------------+-----------+-------------------------------------+
    | **port**           | ``None``  | ``int``                             |
    +--------------------+-----------+-------------------------------------+
    | **url**            | ``None``  | ``str``                             |
    +--------------------+-----------+-------------------------------------+
    
    """
    
    def __init__(self, filer, settings=None):
        self.filer = filer
        self.log = logging.getLogger('pyontapi')
        self.api_modules = {}
        
        if LOG_LEVEL:
            # Set custom loglevel for pyontap logger
            self.log.setLevel(LOG_LEVEL)
        
        self.settings = {
            'user': 'root',
            'password': '',
            'style': constants.LOGIN,
            'vfiler': '',
            'server_type': 'Filer',
            'timeout': 0,
            'transport_type' : constants.HTTP,
            'port': None,
            'url': None,
            'ontapi_version': '1.1',
        }
        
        if settings and isinstance(settings, dict):
            self.__test_settings(settings)
            self.settings.update(settings)
        
        self.__handle_servertype()
        
        if settings and isinstance(settings, dict):
            if not 'transport_type' in settings or \
               settings['transport_type'] == constants.HTTPS:
                # automatically try to use HTTPS if argument not specified or
                # if transport_type is set to HTTPS
                self.__test_https()
        self.__handle_transport_type()
        self.set_api_classes()
    
    def set_api_classes(self):
        """ Add all api classes as class attributes. """
        try:
            result = self.get_ontapi_version()
        except errors.APIFailure, exc:
            self.log.error("Cant't get ontapi version. error was: %s", 
                           exc.get_error())
            raise errors.UnknownOntapiVersionError(exc.errno, 
                    "Cant't get ontapi version. error was:\n%s" % exc.reason
            )
        
        ontapi_version = '%s_%s' % (
            result['major-version'], result['minor-version']
        )
        
        self.settings['ontapi_version'] = '%s.%s' % (
            result['major-version'], result['minor-version']
        )
        path = 'schtob.pyontapi.api.ontapi_%s' % ontapi_version
        try:
            api_module = __import__(path, globals(), locals(), ['API_CLASSES'])
        except ImportError:
            # try to generate api commands
            from schtob.pyontapi.generator import py_gen
            py_gen.generate(self)
            
            try:
                api_module = __import__(path, globals(), locals(), 
                                        ['API_CLASSES']
                )
            except ImportError:
                self.log.error("Can't generate api module for ontapi "
                               "version %s", ontapi_version)
                raise errors.APIGenerationError(-1, "Can't generate api module "
                                "for ontapi version %s" % ontapi_version
                )
        
        for cls in api_module.API_CLASSES:
            if issubclass(cls, api.base.BaseAPI):
                cls.add_to_class(self)
    
    def get_api_modules(self):
        """
        Get all API classes as a dictionary of `package_name`: `api_class`
        """
        return self.api_modules
    
    def get_api_module(self, package_name):
        """ Get API class for ``package_name``. """
        return self.api_modules.get(package_name, None)
    
    def do_api_call(self, api_command_name, arguments, fields):
        """
        Create new API call for ``api_command_name`` using ``arguments`` and 
        return the result as a dictionary using ``fields`` to parse the 
        response.
        """
        
        xmlcontent = self.get_xml_content(api_command_name, arguments)
        if LOG_LEVEL:
            print xmlcontent.toprettyxml()
        try:
            result = self.put_request(xmlcontent)
        except Exception, err:
            raise errors.APIFailure(-1, err.__str__())
        
        if not result.status == 200:
            raise errors.APIFailure(13002, 'HTTP result status "%s"' % 
                             httplib.responses[result.status]
            )
        
        dom = xml.dom.minidom.parseString(result.read())
        if LOG_LEVEL:
            print dom.toprettyxml()
        return self.parse_dom(dom, fields)
    
    def put_request(self, xmlrequest):
        """
        Invoke API command represented in ``xmlrequest`` and return the 
        :class:`httplib.HTTPResponse`.
        """
        
        content = xmlrequest.toxml(encoding='utf-8')
        
        connection = self.get_connection()
        connection.putrequest('POST', self.settings['url'])
        connection.putheader("Content-Length", len(content))
        
        if self.settings['style'] == constants.LOGIN:
            login = self.settings['user']+":"+self.settings['password']
            auth = 'Basic ' + base64.encodestring(login).strip() 
            connection.putheader('Authorization', auth)
        
        connection.endheaders()
        connection.send(content)
        
        response = connection.getresponse()
        connection.close()
        return response
    
    def get_xml_content(self, api_command_name, arguments):
        """
        Create a XML query document for ``api_command_name`` and pass 
        ``arguments``.
        
        ``arguments`` is a list of :class:`schtob.pyontapi.base.NaArgument` 
        instances. Each instance contains the argument value and the 
        information how to generate the corresponding XML structure.
        """
        
        doc = xml.dom.minidom.Document()
        netapp = doc.createElement('netapp')
        
        if self.settings['vfiler']:
            netapp.setAttribute('vfiler', self.settings['vfiler'])
        netapp.setAttribute('version', self.settings['ontapi_version'])
        doc.appendChild(netapp)
        
        api_element = doc.createElement(api_command_name)
        netapp.appendChild(api_element)
        
        for arg in arguments:
            arg.append_to_doc(doc, api_element)
        return doc
    
    def parse_dom(self, dom, fields):
        """
        Parse XML response ``dom`` using ``fields``.
        
        ``fields`` is a list of :class:`schtob.pyontap.api.base.NaField`.
        """
        
        results = dom.getElementsByTagName("results")[0]
        
        status = results.getAttribute('status')
        
        if status != u'passed':
            errno = results.getAttribute('errno')
            reason = results.getAttribute('reason')
            raise errors.APIFailure(errno, reason)
        
        if LOG_LEVEL:
            self.log.debug(results.toprettyxml())
        
        value = {}
        for field in fields:
            value[field.name] = field.get_value(results)
        return value
    
    def get_connection(self):
        """ Returns a HTTP/HTTPS connection instance to filer. """
        
        conn_cls = httplib.HTTPConnection
        if self.settings['transport_type'] == constants.HTTPS:
            conn_cls = httplib.HTTPSConnection
        
        #if sys.version_info >= (2, 6):
        #    return conn_cls(self.filer, self.settings['port'], 
        #                    timeout=self.settings['timeout'])
        return conn_cls(self.filer, self.settings['port'])
    
    def get_ontapi_version(self):
        """
        Invokes API call `system-get-ontapi-version` to get ONTAPI version.
        """
        
        from schtob.pyontapi.generator.system import System
        system_api = System(self)
        return system_api.get_ontapi_version()
    
    #---------------------------PRIVATE METHODS---------------------------------
    
    def __handle_servertype(self):
        """ Set url and port according to ``self.settings['server_type']``. """
        
        server_type = self.settings['server_type']
        update = {
            'port'      : constants.PORTS[server_type],
            'url'       : constants.URLS[server_type],
        }
        self.settings.update(update)
    
    def __test_https(self):
        """ Test if a HTTPS connection is possible for this filer. """
        import socket
        
        server_socket = socket.socket()
        server_socket.settimeout(0.25)
        
        try:
            server_socket.connect((self.filer, 443))
            server_socket.close()
        except socket.error:
            if self.settings['transport_type'] == constants.HTTPS:
                msg = "Fall back to HTTP for filer <%s>; HTTPS was specified!"
                self.log.warning(msg % self.filer)
                self.settings['transport_type'] = constants.HTTP
            return
        self.log.debug('HTTPS test was successful')
        self.settings['transport_type'] = constants.HTTPS
    
    def __handle_transport_type(self):
        """ Set port according to ``self.settings['transport_type'].`` """
        
        if self.settings['transport_type'] == constants.HTTPS:
            self.settings['port'] = 443
    
    def __test_settings(self, settings):
        """ Test if settings are valid. """
        
        test_dict = {
            'style': constants.STYLES,
            'server_type': constants.SERVER_TYPES,
            'transport_type': constants.TRANSPORT_TYPES,
        }
        
        for key, value_list in test_dict.items():
            # Test if kwargs values are valid
            if key in settings and settings[key] not in value_list:
                msg = "%s is not a valid value for %s" % (settings[key], key)
                self.log.error(msg)
                raise ValueError(msg)
